import copy


def empty_field(size=3, empty_sym=' '):
    return [[empty_sym]*size for i in range(size)]


class Position:
    def __init__(self, r, c):
        self.r = r
        self.c = c

    def is_empty_in(self, pf):
        return pf[self] == pf.empty_sym

    def is_inside(self, pf):
        return self.r < len(pf.field) and self.c < len(pf.field[0]) and self.r >= 0 and self.c >= 0

    def __add__(self, other):
        return Position(self.r+other.r, self.c+other.c)

    def __hash__(self):
        return hash((self.r,self.c))

    def __call__(self):
        return self.r, self.c


class Dir:
    def __init__(self, r, c):
        self.r = r
        self.c = c

    def __neg__(self):
        return Dir(-self.r, -self.c)


class PlayField:

    def __init__(self, field, empty_sym, win_length=3):
        self.field = field
        self.empty_sym = empty_sym
        self.win_length = win_length
        self.directions = self.init_directions()
        self.positions_for_distinct_print = []

    def __get__(self):
        return self.field

    def __getitem__(self, pos):
        return self.field[pos.r][pos.c]

    def __str__(self):
        '''
        show the playfield in a form of a simple character grid
        :return: grid as a string
        '''
        size = len(self.field)
        field_as_string = '\n' #empty line as separator
        field_as_string += "  "
        # preparing header - top row
        for c in range(size):
            field_as_string += ' %d'%c
        field_as_string += "\n"
        field_as_string += ' |'
        for c in range(size):
            field_as_string += '--'
        field_as_string += "\n"

        for r in range(size): # for each row
            field_as_string += '%d|'%r # row id
            for c in range(size): # for all columns
                # adding the field content
                if (r,c) in self.positions_for_distinct_print:
                    field_as_string += ' '+self.field[r][c].upper()
                else:
                    field_as_string += ' '+self.field[r][c]
            field_as_string += '\n'
        return(field_as_string)

    def init_directions(self):
        dirs = []
        for dir in [(0,1), (1, 1), (1, 0), (1, -1)]:
            dirs.append(Dir(dir[0], dir[1]))
        return dirs

    def is_winning(self):
        for pos in self.non_empty_pos():
            sequences = self.get_all_sequences(pos, self[pos])
            for seq, length in sequences.items():
                if length == self.win_length:
                    return True, seq
        return False, None

    def get_all_sequences(self, pos, sym):
        """
        :param pos: Position object
        :param sym: symbol of the player that moves
        :return: dictionary: sequences of positions as the keys and their lengths as values
        """
        sequences = {}
        for direction in self.directions: # forward
            forward_seq = self.get_seq(pos, direction, sym)
            backward_seq = self.get_seq(pos, -direction, sym)
            complete_sequence = forward_seq | backward_seq
            sequences[complete_sequence] = len(complete_sequence)
        return sequences

    def get_seq(self, start_pos, direction, sym):
        positions = [start_pos]
        pos = start_pos + direction
        while pos.is_inside(self) and self[pos]==sym:
            positions.append(pos)
            pos = pos+direction
        return frozenset(positions)

    def get_empty_positions(self):
        return list(self.empty_pos())

    def is_full(self):
        for pos in self.empty_pos():
            return False  # one empty position is enough
        return True

    def update(self, move, sym):
        self.field[move[0]][move[1]]= sym

    def copy_field(self):
        return copy.deepcopy(self.field)

    def show_sequence(self,seq):
        for pos in seq:
            self.positions_for_distinct_print.append(pos())
        print(self)

    # auxiliary generators
    def empty_pos(self):
        for pos in self.all_pos():
            if pos.is_empty_in(self):
                yield pos

    def all_pos(self):
        for r in range(len(self.field)):
            for c in range(len(self.field[0])):
                yield Position(r, c)

    def non_empty_pos(self):
        for pos in self.all_pos():
            if not(pos.is_empty_in(self)):
                yield pos
